@@ -23,6 +23,10 @@ class TokenType(enum.Enum):
     SEMICOLON = enum.auto()
     DOT = enum.auto()
     NOT_EQ = enum.auto()
+    GREATER_THAN = enum.auto()
+    LESS_THAN = enum.auto()
+    GREATER_THAN_OR_EQ = enum.auto()
+    LESS_THAN_OR_EQ = enum.auto()
     STRING = enum.auto()
     BLOB = enum.auto()
     INTEGER = enum.auto()
@@ -58,12 +62,12 @@ class Lexer:
           - A (TokenType, string) pair, which matches a token with that type and that
               value.
         """
-        if self.done():
+        token = self.current()
+        if token is None:
             raise SQLiteParserError("premature end of input")
 
-        token = self.current()
         if token.type == TokenType.UNKNOWN:
-            raise SQLiteParserError("unknown token")
+            raise SQLiteParserError(f"unknown token: {token.value!r}")
 
         for type_or_value in types_and_values:
             if isinstance(type_or_value, str):
@@ -112,9 +116,15 @@ class Lexer:
         elif c.isdigit():
             return self.read_integer()
         elif c == "'":
-            return self.read_string()
+            return self.read_generic_string(TokenType.STRING, "'")
         elif c == '"':
-            return self.read_quoted_symbol()
+            return self.read_generic_string(TokenType.IDENTIFIER, '"')
+        elif c == "`":
+            return self.read_generic_string(TokenType.IDENTIFIER, "`")
+        elif c == "[":
+            return self.read_generic_string(
+                TokenType.IDENTIFIER, "]", allow_doubling=False
+            )
         elif c == "(":
             return self.character_token(TokenType.LEFT_PARENTHESIS)
         elif c == ")":
@@ -127,6 +137,14 @@ class Lexer:
             return self.character_token(TokenType.DOT)
         elif self.prefix(2) == "!=":
             return self.multi_character_token(TokenType.NOT_EQ, 2)
+        elif self.prefix(2) == ">=":
+            return self.multi_character_token(TokenType.GREATER_THAN_OR_EQ, 2)
+        elif self.prefix(2) == "<=":
+            return self.multi_character_token(TokenType.LESS_THAN_OR_EQ, 2)
+        elif c == ">":
+            return self.character_token(TokenType.GREATER_THAN)
+        elif c == "<":
+            return self.character_token(TokenType.GREATER_THAN)
         else:
             return self.character_token(TokenType.UNKNOWN)
 
@@ -212,13 +230,7 @@ class Lexer:
             column=start_column,
         )
 
-    def read_string(self):
-        return self.read_generic_string(TokenType.STRING, "'")
-
-    def read_quoted_symbol(self):
-        return self.read_generic_string(TokenType.IDENTIFIER, '"')
-
-    def read_generic_string(self, token_type, delimiter):
+    def read_generic_string(self, token_type, delimiter, *, allow_doubling=True):
         start_column = self.column
         characters = []
 
@@ -226,7 +238,7 @@ class Lexer:
         while not self.done():
             if self.c() == delimiter:
                 # SQL escapes quotes by doubling them.
-                if self.prefix(2) == delimiter * 2:
+                if allow_doubling and self.prefix(2) == delimiter * 2:
                     characters.append(delimiter)
                     self.next_character()
                     self.next_character()
@@ -261,6 +273,9 @@ class Lexer:
         return self.prefix(1)
 
     def peek(self, n=1):
+        if self.index + n >= len(self.program):
+            return None
+
         return self.program[self.index + n]
 
     def prefix(self, length):
